try:
    import simplegui
except ImportError:
    import SimpleGUICS2Pygame.simpleguics2pygame as simplegui

import math, random

from Vector import Vector

# Entity file is for all living and interactable entities
# Player Constants in Player
# Enemy Constants in Enemy 

class Entity:
    """Parent class for all generated living entities in the game."""
    # Sets som default values but allows for other values to be passed in as well
    def __init__(self, position, radius, color, sprite_url=None, sprite_rows=1, sprite_columns=1):
        self.position = position
        self.radius = radius
        self.color = color
        self.velocity = Vector(0, 0)
        self.health = 0
        self.max_health = 0
        self.alive = True

        # Sprite properties
        self.sprite_url = sprite_url
        self.sprite_rows = sprite_rows
        self.sprite_columns = sprite_columns

        if sprite_url:
            self.image = simplegui.load_image(sprite_url)
            self.img_width = self.image.get_width()
            self.img_height = self.image.get_height()
            self.frame_width = self.img_width // sprite_columns
            self.frame_height = self.img_height // sprite_rows
            self.current_frame = 0
            self.sprite_row = 0  # Default to first row
            self.animation_counter = 0
            self.ANIMATION_SPEED = 15  # Default animation speed

    def update(self):
        self.position.add(self.velocity)
    
    def draw(self, canvas):
        """Draw entity sprite if available, otherwise draw a circle."""
        if self.sprite_url:
            self.draw_sprite(canvas)
        else:
            pos = self.position.get_p()
            canvas.draw_circle(pos, self.radius, 1, "Black", self.color)
    
    def collides_with(self, other):
        """Check if this entity collides with another entity object. Used

        Used to determine if enemy object is in contact with player.
        
        Args:
            other (Entity): The other entity to check collision with
            
        Returns:
            bool: True if entities are colliding, False otherwise
        """
        distance = (self.position - other.position).length()
        return distance < self.radius + other.radius
    
    def draw_sprite(self, canvas):
        """Handles drawing the sprite from the sprite sheet."""
        frame_center = (
            (self.current_frame + 0.5) * self.frame_width,
            (self.sprite_row + 0.5) * self.frame_height
        )
        frame_size = (self.frame_width, self.frame_height)
        
        canvas.draw_image(self.image, frame_center, frame_size,
                          self.position.get_p(), (self.radius * 3.5, self.radius * 3.5))
    
    def take_damage(self, damage):
        self.health -= damage
        if self.health <= 0:
            self.alive = False

class Player(Entity):
    """Player class, controlled by user."""

    # Player constants
    PLAYER_RADIUS = 20
    PLAYER_COLOR = "Blue"
    PLAYER_MAX_HEALTH = 100
    PLAYER_INVULNERABILITY_DURATION = 60  # in frames after being hit

    SPRITE_URL = "https://www.cs.rhul.ac.uk/home/znac612/cs1822/player.png"
    SPRITE_ROWS = 4
    SPRITE_COLUMNS = 15
    WALKING_SPRITE_COLUMNS = 9
    
    def __init__(self, position):
        super().__init__(position, self.PLAYER_RADIUS, self.PLAYER_COLOR, 
                         self.SPRITE_URL, self.SPRITE_ROWS, self.SPRITE_COLUMNS)
        
        self.speed = 5
        self.health = self.PLAYER_MAX_HEALTH
        self.max_health = self.PLAYER_MAX_HEALTH
        self.attack_radius = 25
        self.attack_damage = 25

        self.attack_duration = 10  # in frames
        self.attack_cooldown = 30  # in frames
        self.is_attacking = False
        self.attack_timer = 0
        self.attack_cooldown_timer = 0
        
        self.invulnerability_duration = self.PLAYER_INVULNERABILITY_DURATION
        self.is_invulnerable = False
        self.invulnerable_timer = 0

        self.animation_speed = 5
        self.last_sprite_row = 3
        self.state = "walking"
        self.is_in_attack_animation = False

        self.direction = Vector(1, 0)  # initially facing right

        self.on_damage = None

        self.key_states = {
            'w': False,
            'a': False,
            's': False,
            'd': False
        }
    
    def key_down(self, key):
        if key == simplegui.KEY_MAP['w']:
            self.key_states['w'] = True
        elif key == simplegui.KEY_MAP['a']:
            self.key_states['a'] = True
        elif key == simplegui.KEY_MAP['s']:
            self.key_states['s'] = True
        elif key == simplegui.KEY_MAP['d']:
            self.key_states['d'] = True
        elif key == simplegui.KEY_MAP['space']:
            self.attack()
    
    def key_up(self, key):
        if key == simplegui.KEY_MAP['w']:
            self.key_states['w'] = False
        elif key == simplegui.KEY_MAP['a']:
            self.key_states['a'] = False
        elif key == simplegui.KEY_MAP['s']:
            self.key_states['s'] = False
        elif key == simplegui.KEY_MAP['d']:
            self.key_states['d'] = False

    def update(self):
        direction_x, direction_y = 0, 0
        if self.key_states['a']: direction_x -= 1
        if self.key_states['d']: direction_x += 1
        if self.key_states['w']: direction_y -= 1
        if self.key_states['s']: direction_y += 1

        # Normalize diagonal movement
        if direction_x != 0 and direction_y != 0:
            length = math.sqrt(direction_x * direction_x + direction_y * direction_y)
            direction_x /= length
            direction_y /= length

        self.set_velocity(direction_x, direction_y)
        super().update()

        # Determine movement direction using dominant axis
        dx = self.velocity.x
        dy = self.velocity.y

        # Attack animation plays as long as attack is on cool down
        if self.is_attacking or self.attack_cooldown_timer > 0:
            self.state = "attacking"
        elif dx != 0 or dy != 0:
            self.state = "walking"
        else:
            self.state = "idle"

        # Set sprite row based on movement
        if abs(dx) > abs(dy):  
            self.sprite_row = 3 if dx > 0 else 1  # Right / Left
        elif dx != 0 or dy != 0:
            self.sprite_row = 2 if dy > 0 else 0  # Down / Up

        # If idle, retain the last movement direction
        if dx == 0 and dy == 0:
            self.sprite_row = self.last_sprite_row
        else:
            self.last_sprite_row = self.sprite_row

        # Update animation frame
        self.animation_counter += 1
        if self.animation_counter >= self.animation_speed:
            self.animation_counter = 0
            if self.state == "walking":
                self.current_frame = (self.current_frame + 1) % self.WALKING_SPRITE_COLUMNS
            elif self.state == "attacking":
                if self.current_frame < self.WALKING_SPRITE_COLUMNS:
                    self.current_frame = self.WALKING_SPRITE_COLUMNS
                if self.current_frame < 14:  # Doesn't loop animation to look cleaner
                    self.current_frame += 1
            elif self.state == "idle":
                self.current_frame = 0  # Stay on idle frame

        # Handle attack mechanics 
        # Attck and attack animation have different timings
        if self.is_attacking:
            self.attack_timer -= 1
            if self.attack_timer <= 0:
                self.is_attacking = False  # The actual attack ends
                self.attack_cooldown_timer = self.attack_cooldown  # Start cooldown

        # Reduce attack cooldown timer
        if self.attack_cooldown_timer > 0:
            self.attack_cooldown_timer -= 1

        # Handle invulnerability timer
        if self.is_invulnerable:
            self.invulnerable_timer -= 1
            if self.invulnerable_timer <= 0:
                self.is_invulnerable = False

   
    def draw(self, canvas):
        super().draw(canvas)

        if self.is_attacking:
            attack_pos = self.position + self.direction.get_normalized() * (self.attack_radius + 10)

            
            canvas.draw_line(self.position.get_p(), attack_pos.get_p(), 4, "Orange")
            
            canvas.draw_line(self.position.get_p(), attack_pos.get_p(), 2, "Yellow")

            
            canvas.draw_circle(attack_pos.get_p(), 6, 1, "Orange", "Red")
            canvas.draw_circle(attack_pos.get_p(), 3, 1, "Yellow", "Yellow")

            
            for i in range(3):
                spark_offset = self.direction.get_normalized().multiply(5 + i * 10)
                spark_pos = self.position + spark_offset
                spark_pos.x += random.randint(-3, 3)
                spark_pos.y += random.randint(-3, 3)
                canvas.draw_circle(spark_pos.get_p(), 2, 1, "Red", "Orange")


    
    def set_velocity(self, x, y):
        self.velocity = Vector(x, y).multiply(self.speed)
        
        # Update direction when moving
        if x != 0 or y != 0:
            self.direction = Vector(x, y).get_normalized()
    
    def attack(self):
        """Try to perform an attack if not on cooldown.
        
        Returns:
            bool: True if attack was started, False if on cooldown
        """
        if not self.is_attacking and self.attack_cooldown_timer <= 0:
            self.is_attacking = True
            self.is_in_attack_animation = True  # Keep attack animation going
            self.attack_timer = self.attack_duration
            return True
        return False
    
    def take_damage(self, damage):
        if not self.is_invulnerable:
            super().take_damage(damage)
            self.is_invulnerable = True
            self.invulnerable_timer = self.invulnerability_duration

            if self.on_damage:
                self.on_damage()

class Enemy(Entity):
    """Enemy entity which in its current behaviour, only chases and attacks the player.
    
    Enemies move toward the player and deal damage on contact.
    """

    # Enemy constants
    ENEMY_RADIUS = 15
    ENEMY_COLOR = "Red"
    ENEMY_SPEED = 2
    ENEMY_MIN_HEALTH = 200
    ENEMY_MAX_HEALTH = 350
    ENEMY_DAMAGE = 10
    ENEMY_ATTACK_COOLDOWN = 60  # in frames
    SPRITE_ROWS = 5
    SPRITE_COLUMNS = 11  # Each row has 11 frames
    WALKING_SPRITE_COLUMNS = 6 # Only walking, not attacking
    ANIMATION_SPEED = 15  # Change frame every 15 updates
    ENEMY_SPRITE = "https://www.cs.rhul.ac.uk/home/znac612/cs1822/enemy.png"

    def __init__(self, position):
        super().__init__(position, self.ENEMY_RADIUS, self.ENEMY_COLOR, 
                         self.ENEMY_SPRITE,
                         self.SPRITE_ROWS, self.SPRITE_COLUMNS)
        
        self.health = random.randint(self.ENEMY_MIN_HEALTH, self.ENEMY_MAX_HEALTH)
        self.max_health = self.health
        self.attack_timer = 0
        self.speed = self.ENEMY_SPEED
        self.damage = self.ENEMY_DAMAGE
        self.attack_cooldown = self.ENEMY_ATTACK_COOLDOWN
        self.state = "walking" # Either walking or attacking

    def update(self, player_pos):
        # Move towards player
        direction_to_player = player_pos - self.position
        if direction_to_player.length() > 0:
            self.velocity = direction_to_player.get_normalized().multiply(self.speed)
        super().update()

        # Determine movement direction using dominant axis
        dx = self.velocity.x
        dy = self.velocity.y
        if abs(dx) > abs(dy):  
            if dx > 0:
                self.sprite_row = 1  # Right (1)
            else:
                self.sprite_row = 3  # Left (3)
        else:
            if dy > 0:
                self.sprite_row = 0  # Down (0)
            else:
                self.sprite_row = 2  # Up (2)

        # Update animation frame
        self.animation_counter += 1
        if self.animation_counter >= self.ANIMATION_SPEED:
            self.animation_counter = 0
            if self.state == "walking":
                self.current_frame = (self.current_frame + 1) % self.WALKING_SPRITE_COLUMNS
            elif self.state == "attacking":
                if self.current_frame < 7 or self.current_frame >= 9:
                    self.current_frame = 7  # Reset to first attack frame
                else:
                    self.current_frame += 1  # Cycle through attack frames
        
        if self.attack_timer > 0:
            self.attack_timer -= 1

        if self.attack_timer <= 0:
            self.state = "walking"

    def draw(self, canvas):
        """Draws healthbar on sprite and draw image over the enemy"""
        
        health_percentage = self.health / self.max_health
        health_width = self.radius * 2 * health_percentage
        
        health_bar_y = self.position.y - self.radius - 8
        health_bar_x = self.position.x - self.radius
        
        # Background of health bar
        canvas.draw_line(
            (health_bar_x, health_bar_y),
            (health_bar_x + self.radius * 2, health_bar_y),
            3, "White"
        )
        
        # Current health status
        canvas.draw_line(
            (health_bar_x, health_bar_y),
            (health_bar_x + health_width, health_bar_y),
            3, "Red"
        )

        # Draw sprite using Entity's method
        super().draw(canvas)
   
    def attack(self, player):
        """Attack the player if in range and not on cooldown.

        Returns:
            bool: True if attack was performed, False otherwise
        """
        if self.attack_timer <= 0 and self.collides_with(player):
            self.state = "attacking"
            player.take_damage(self.damage)
            self.attack_timer = self.attack_cooldown
            return True
        return False
    
    def draw_sprite(self, canvas):
        """Need to change the sprite sheet rendering as the enemy sprite has an odd offset"""
        frame_center = (
            (self.current_frame + 0.5) * self.frame_width + self.current_frame,
            (self.sprite_row + 0.5) * self.frame_height
        )
        frame_size = (self.frame_width, self.frame_height)
        
        canvas.draw_image(self.image, frame_center, frame_size,
                          self.position.get_p(), (self.radius * 3.5, self.radius * 3.5))

class Obstacle(Entity):
    """Rectanglular obstacles block the movement of entities.
    
    Attributes:
        width (float): Width of the obstacle
        height (float): Height of the obstacle
        bounds (dict): Boundary coordinates object can spawn in (left, right, top, bottom) 
    """

    def __init__(self, position, width, height):
        super().__init__(position, 0, "Gray")
        self.width = width
        self.height = height
        self.bounds = {
            "left": position.x - width/2,
            "right": position.x + width/2,
            "top": position.y - height/2,
            "bottom": position.y + height/2
        }
    
    def draw(self, canvas):
        top_left = (self.bounds["left"], self.bounds["top"])
        bottom_right = (self.bounds["right"], self.bounds["bottom"])
        canvas.draw_polygon([
            top_left,
            (self.bounds["right"], self.bounds["top"]),
            bottom_right,
            (self.bounds["left"], self.bounds["bottom"])
        ], 1, "Black", self.color)
        
    def collides_with_circle(self, center, radius):
        """Check if player collides with obstacle.
        
        Args:
            center (Vector): The center point of the player
            radius (float): The radius of the player
            
        Returns:
            bool: True if colliding, False otherwise
        """
        # Find the closest point on rectangle to the circle center
        closest_x = max(self.bounds["left"], min(center.x, self.bounds["right"]))
        closest_y = max(self.bounds["top"], min(center.y, self.bounds["bottom"]))
        
        # Calculate distance from that point to circle center
        distance = math.sqrt((closest_x - center.x) ** 2 + (closest_y - center.y) ** 2)
        
        return distance < radius

class Item(Entity):
    """ Acts like a live entity, must be attacked to gain the power up.
    When one item is taken from a rooom all others dissapear"""

    # Item Constants
    # Fixed positions for the items in a room
    ITEM_POS_1 = [(100, 100), (120, 100), (120, 120), (100, 120)]
    ITEM_POS_2 = [(400, 100), (420, 100), (420, 120), (400, 120)]
    ITEM_POS_3 = [(700, 100), (720, 100), (720, 120), (700, 120)]

    ITEM_RADIUS = 10

    # Chances for the rarity of an item
    COMMON_CHANCE = 50 # 50% chance
    UNCOMMON_CHANCE = 75 # 25% chance
    RARE_CHANCE = 88 # 13% chance
    EPIC_CHANCE = 96 # 8% chance
    # Legendary is 4% chance

    # Various stats the player can gain
    POWER_UPS = ["SPEED", 
                    "MAX HP",
                    "ATK SIZE", 
                    "ATK SPD", 
                    "DAMAGE"]

    def __init__(self, number):
        # Takes in a number to determine which item in the room this instance is
        # Choose a rarity and stat to modify
        self.rarity = self.pickRarity()
        self.powerUp = self.pickPowerUp()

        # Used to determine which position this item should be drawn at
        self.number = number
        
        if self.number == 1:
            self.points = self.ITEM_POS_1
        elif self.number == 2:
            self.points = self.ITEM_POS_2
        elif self.number == 3:
            self.points = self.ITEM_POS_3

        self.pos = Vector((self.points[0][0] + self.points[2][0]) / 2, (self.points[0][1] + self.points[2][1]) / 2)

        super().__init__(self.pos, self.ITEM_RADIUS, self.color)
        
        # Text above item to say what it is
        self.description = self.rarity + " : " + self.powerUp
    
    def pickRarity(self):
        # Determine how good the item is
        # Affects the colour of the item, how much the stat increase is
        roll = random.randint(0,100)
        if roll < self.COMMON_CHANCE:
            self.color = "Grey"
            return "Common"
        elif roll < self.UNCOMMON_CHANCE:
            self.color = "Green"
            return "Uncommon"
        elif roll < self.RARE_CHANCE:
            self.color = "Blue"
            return "Rare"
        elif roll < self.EPIC_CHANCE:
            self.color = "Purple"
            return "Epic"
        else:
            self.color = "Yellow"
            return "Legendary"
        
    def pickPowerUp(self):
        return self.POWER_UPS[random.randint(0,4)]
    
    def draw(self, canvas):
        # Draw Box for item
        canvas.draw_polygon(self.points, 1, "Black", self.color)
        
        text_pos = (self.pos.x - 50, self.pos.y - self.ITEM_RADIUS - 10)

        # Draw background for text to make it readable
        canvas.draw_line(
            (text_pos[0] - 5, text_pos[1] - 6),
            (text_pos[0] + self.radius * 9, text_pos[1]),
            20, "Black"
        )

        # Draw the item description above the item
        canvas.draw_text(self.description, text_pos, 10, self.color)
    
    def add_powerUp(self, player):

        power_up = self.powerUp

        # Define stat increases based on rarity
        rarity_multipliers = {
            "Common": 1,
            "Uncommon": 1.5,
            "Rare": 2,
            "Epic": 3,
            "Legendary": 5
        }

        multiplier = rarity_multipliers[self.rarity] 

        # Various stats have different amounts to increase by to keep game balanced
        if power_up == "SPEED":
            increase = int(0.5 * multiplier)
            player.speed += increase
        elif power_up == "MAX HP":
            increase = int(10 * multiplier)
            player.max_health += increase
            player.health = min(player.health + increase, player.max_health)
        elif power_up == "ATK SIZE":
            increase = int(5 * multiplier)
            player.attack_radius += increase
        elif power_up == "ATK SPD":
            increase = round(0.2 * multiplier, 2)
            player.attack_cooldown = max(0.5, player.attack_cooldown - increase)
        elif power_up == "DAMAGE":
            increase = int(10 * multiplier)
            player.attack_damage += increase

class BossEnemy(Enemy):

    # Boss Types
    BOSS_TYPES = [
        {  # High health, low speed, big damage
            "color": "Pink",
            "health": 1500,
            "speed": 1,
            "damage": 40,
            "radius": 85
        },
        {  # Low health, high speed, low damage
            "color": "Yellow",
            "health": 750,
            "speed": 3.5,
            "damage": 15,
            "radius": 60
        },
        {  # High speed, high health, low damage
            "color": "Orange",
            "health": 900,
            "speed": 2.25,
            "damage": 20,
            "radius": 40
        }
    ]

    BOSS_ATTACK_COOLDOWN = 90  # in frames

    SPRITE_ROWS = 4
    SPRITE_COLUMNS = 14  # Each row has 14 frames
    WALKING_SPRITE_COLUMNS = 6 # Only walking, not attacking
    ANIMATION_SPEED = 15  # Change frame every 15 updates
    BOSS_SPRITE = "https://www.cs.rhul.ac.uk/home/znac612/cs1822/boss.png"

    def __init__(self, position):
        boss_type = random.choice(self.BOSS_TYPES)
        
        Enemy.__init__(self, position)
        
        self.radius = boss_type["radius"]
        self.color = boss_type["color"]
        self.health = boss_type["health"]
        self.max_health = boss_type["health"]
        self.damage = boss_type["damage"]
        self.speed = boss_type["speed"]
        self.attack_cooldown = self.BOSS_ATTACK_COOLDOWN
        self.attack_timer = 0
        
        self.image = simplegui.load_image(self.BOSS_SPRITE)
        self.img_width = self.image.get_width()
        self.img_height = self.image.get_height()
        self.frame_width = self.img_width // self.SPRITE_COLUMNS
        self.frame_height = self.img_height // self.SPRITE_ROWS
        
    def update(self, player_pos):
        # Move towards player
        direction_to_player = player_pos - self.position
        if direction_to_player.length() > 0:
            self.velocity = direction_to_player.get_normalized().multiply(self.speed)
        Entity.update(self)

        # Determine movement direction using dominant axis
        dx = self.velocity.x
        dy = self.velocity.y
        if abs(dx) > abs(dy):  
            if dx > 0:
                self.sprite_row = 3  # Right (3)
            else:
                self.sprite_row = 1  # Left (1)
        else:
            if dy > 0:
                self.sprite_row = 2  # Down (2)
            else:
                self.sprite_row = 0  # Up (0)

        # Update animation frame
        self.animation_counter += 1
        if self.animation_counter >= self.ANIMATION_SPEED:
            self.animation_counter = 0
            if self.state == "walking":
                self.current_frame = (self.current_frame + 1) % self.WALKING_SPRITE_COLUMNS
            elif self.state == "attacking":
                if self.current_frame < 8 or self.current_frame >= 13:
                    self.current_frame = 8  # Reset to first attack frame
                else:
                    self.current_frame += 1  # Cycle through attack frames
        
        if self.attack_timer > 0:
            self.attack_timer -= 1

        if self.attack_timer <= 0:
            self.state = "walking"

    def draw_sprite(self, canvas):
        Entity.draw_sprite(self, canvas)