try:
    import simplegui
except ImportError:
    import SimpleGUICS2Pygame.simpleguics2pygame as simplegui

import math

from Vector import Vector
from Entity import Player
from Room import Room, EnemyRoom, BossRoom, ItemRoom

# Interaction file is for all game functions and interactions
# HUD constants

class Interaction:
    """Handle all game interactions.
    
    Manages the overall game state, game loop, controls and flow of game.
    """

    # Main Menu Image
    MAIN_MENU = simplegui.load_image("https://www.cs.rhul.ac.uk/home/znac612/cs1822/mainMenu.png")
    FINISHED = simplegui.load_image("https://www.cs.rhul.ac.uk/home/znac120/cs1822/GameComplete.png")
    FLOOR = simplegui.load_image("https://www.cs.rhul.ac.uk/home/znac612/cs1822/floor.png")

    # Play Button Constants
    PLAY_BUTTON_WIDTH = 100
    PLAY_BUTTON_HEIGHT = 40
    PLAY_BUTTON_X = Room.CANVAS_WIDTH / 2 - PLAY_BUTTON_WIDTH / 2
    PLAY_BUTTON_Y = Room.CANVAS_HEIGHT / 2 + 100

    # HUD constants
    HUD_PADDING = 10
    HUD_HEIGHT = 30
    HUD_BG_COLOR = "Gray" # if want to change design in the future
    HUD_TEXT_COLOR = "White"
    HUD_HEALTH_COLOR = "Red"

    def __init__(self):
        self.state = "menu" # Either "menu" or "game" or "finished"
        self.player = Player(Vector(Room.CANVAS_WIDTH/2, Room.CANVAS_HEIGHT - Room.WALL_THICKNESS - 40))
        self.player.on_damage = lambda: self.trigger_flash()
        self.room_count = 1
        self.room = EnemyRoom(self.room_count)
        self.boss_room_count = 0 # Count boss rooms to finish game at 3
        self.game_over = False
        self.score = 0
        self.flash_timer = 0
        self.lives = 3
        
        self.key_states = {
            'w': False,
            'a': False,
            's': False,
            'd': False
        }
    
    def update(self):
        if self.game_over or self.state == "menu" or self.state == "finished":
            return
        
        self.player.update()

        self.room.update(self.player, self) # Room being called needs its own update method
        self.room.handle_collisions(self.player)
        
        if not self.player.alive:
            if self.lives > 1:
                self.lives -= 1
                self.player = Player(Vector(Room.CANVAS_WIDTH/2, Room.CANVAS_HEIGHT - Room.WALL_THICKNESS - 40))
                self.player.on_damage = lambda: self.trigger_flash()
            else:  
                self.game_over = True

        if self.flash_timer > 0:
            self.flash_timer -= 1 

        # Check for room transition or finishing game
        door_entered = self.room.check_door_entered(self.player)

        # Check if all conditions for winning are fufilled
        # 3 bosses killed
        # Door open / Killed current boss
        # Entered final door
        if self.boss_room_count >= 3 and self.room.door[0].is_open and (door_entered != -1):
            self.state = "finished"

        if door_entered != -1:
            self.new_room()   
    
    def new_room(self):
        self.room_count += 1
        if self.room_count % 5 == 0:  # Boss room every 5th room
            self.room = BossRoom(self.room_count)
            self.boss_room_count += 1
        elif self.room_count % 3 == 0: # Item room every 3rd room
            self.room = ItemRoom()
        else:
            self.room = EnemyRoom(self.room_count)
        
        # Position player at the bottom center of the room
        self.player.position = Vector(Room.CANVAS_WIDTH/2, Room.CANVAS_HEIGHT - Room.WALL_THICKNESS - 2*self.player.radius)
    
    def draw(self, canvas):

        if self.state == "menu":
            self.draw_main_menu(canvas)
        elif self.state == "finished":
            self.draw_finished(canvas)
        else:
            canvas.draw_image(self.FLOOR,
                        (self.FLOOR.get_width() / 2, self.FLOOR.get_height() / 2),  # Center of image
                        (self.FLOOR.get_width(), self.FLOOR.get_height()),          # Image size
                        (Room.CANVAS_WIDTH / 2, Room.CANVAS_HEIGHT / 2),            # Draw position
                        (Room.CANVAS_WIDTH, Room.CANVAS_HEIGHT))                    # Scale to fit
            self.room.draw(canvas)
            self.player.draw(canvas)
            if not self.game_over:
                self.draw_hud(canvas)
        
            if self.flash_timer > 0:
                canvas.draw_polygon(
                [(0, 0), (Room.CANVAS_WIDTH, 0), (Room.CANVAS_WIDTH, Room.CANVAS_HEIGHT), (0, Room.CANVAS_HEIGHT)],
                1, "Red", "rgba(255, 0, 0, 0.4)"
                )

            if self.game_over:
                canvas.draw_text("GAME OVER", (Room.CANVAS_WIDTH/2 - 100, Room.CANVAS_HEIGHT/2), 40, "Red")
                canvas.draw_text("Your score: " + str(self.score), (Room.CANVAS_WIDTH/2 - 80, Room.CANVAS_HEIGHT/2 + 40), 20, "White")
                canvas.draw_text("Press R to restart", (Room.CANVAS_WIDTH/2 - 80, Room.CANVAS_HEIGHT/2 + 70), 20, "White")
    
    def trigger_flash(self):
        self.flash_timer = 10

    def draw_hud(self, canvas):
        canvas.draw_polygon([
            (self.HUD_PADDING, self.HUD_PADDING),
            (Room.CANVAS_WIDTH/3, self.HUD_PADDING),
            (Room.CANVAS_WIDTH/3, self.HUD_PADDING + self.HUD_HEIGHT),
            (self.HUD_PADDING, self.HUD_PADDING + self.HUD_HEIGHT)
        ], 1, "Black", self.HUD_BG_COLOR)
        
        health_percentage = self.player.health / self.player.max_health
        health_width = (Room.CANVAS_WIDTH/3 - self.HUD_PADDING*2) * health_percentage
        
        canvas.draw_polygon([
            (self.HUD_PADDING*2, self.HUD_PADDING*2),
            (self.HUD_PADDING*2 + health_width, self.HUD_PADDING*2),
            (self.HUD_PADDING*2 + health_width, self.HUD_PADDING + self.HUD_HEIGHT - self.HUD_PADDING),
            (self.HUD_PADDING*2, self.HUD_PADDING + self.HUD_HEIGHT - self.HUD_PADDING)
        ], 1, "Black", self.HUD_HEALTH_COLOR)
        
        canvas.draw_text(f"HP: {self.player.health}/{self.player.max_health}", 
                        (self.HUD_PADDING*2 + 5, self.HUD_PADDING + self.HUD_HEIGHT/2 + 5), 
                        12, self.HUD_TEXT_COLOR)
        
        canvas.draw_text(f"Room: {self.room_count}", 
                        (Room.CANVAS_WIDTH/3 - 80, self.HUD_PADDING + self.HUD_HEIGHT/2 + 5), 
                        12, self.HUD_TEXT_COLOR)
        
        # Door status
        door_status = "OPEN" if self.room.door[0].is_open else "LOCKED"
        door_color = "Green" if self.room.door[0].is_open else "Red"
        canvas.draw_text(f"Door: {door_status}", 
                        (Room.CANVAS_WIDTH - 100, self.HUD_PADDING + self.HUD_HEIGHT/2 + 5), 
                        12, door_color)
        
        # Enemies remaining
        if isinstance(self.room, EnemyRoom):
            enemies_remaining = len(self.room.enemies)
        else:
            enemies_remaining = 0
        canvas.draw_text(f"Enemies: {enemies_remaining}", 
                        (Room.CANVAS_WIDTH - 200, self.HUD_PADDING + self.HUD_HEIGHT/2 + 5), 
                        12, self.HUD_TEXT_COLOR)
    

        canvas.draw_text(f"Score: {self.score}", 
                         (Room.CANVAS_WIDTH - 300, self.HUD_PADDING + self.HUD_HEIGHT/2 + 5), 
                         12, self.HUD_TEXT_COLOR)

        canvas.draw_text(f"Lives: {self.lives}", 
                 (Room.CANVAS_WIDTH - 380, self.HUD_PADDING + self.HUD_HEIGHT/2 + 5), 
                 12, self.HUD_TEXT_COLOR)                 
                         
    def key_down(self, key):
        self.player.key_down(key)
        if key == simplegui.KEY_MAP['r'] and self.game_over:
            self.reset_game()
    
    def key_up(self, key):
        self.player.key_up(key)
    
    def reset_game(self):
        """Reset the game to initial state after game over."""
        self.player = Player(Vector(Room.CANVAS_WIDTH/2, Room.CANVAS_HEIGHT - Room.WALL_THICKNESS - 40))
        self.room_count = 1
        self.room = EnemyRoom(self.room_count)
        self.boss_room_count = 0
        self.game_over = False
        self.lives = 3
        self.score = 0
        self.state = "menu"

    def draw_main_menu(self, canvas):
        canvas.draw_image(self.MAIN_MENU,
                        (self.MAIN_MENU.get_width() / 2, self.MAIN_MENU.get_height() / 2),  # Center of image
                        (self.MAIN_MENU.get_width(), self.MAIN_MENU.get_height()),          # Image size
                        (Room.CANVAS_WIDTH / 2, Room.CANVAS_HEIGHT / 2),                    # Draw position
                        (Room.CANVAS_WIDTH, Room.CANVAS_HEIGHT))                            # Scale to fit

        # Draw Play button
        canvas.draw_polygon([(self.PLAY_BUTTON_X, self.PLAY_BUTTON_Y),
                            (self.PLAY_BUTTON_X + self.PLAY_BUTTON_WIDTH, self.PLAY_BUTTON_Y),
                            (self.PLAY_BUTTON_X + self.PLAY_BUTTON_WIDTH, self.PLAY_BUTTON_Y + self.PLAY_BUTTON_HEIGHT),
                            (self.PLAY_BUTTON_X, self.PLAY_BUTTON_Y + self.PLAY_BUTTON_HEIGHT)],
                            2, "White", "Gray"
                            )

        canvas.draw_text("PLAY", (self.PLAY_BUTTON_X + 27, self.PLAY_BUTTON_Y + 30), 20, "White")

    def draw_finished(self, canvas):
        canvas.draw_image(self.FINISHED,
                        (self.FINISHED.get_width() / 2, self.FINISHED.get_height() / 2),  # Center of image
                        (self.FINISHED.get_width(), self.FINISHED.get_height()),          # Image size
                        (Room.CANVAS_WIDTH / 2, Room.CANVAS_HEIGHT / 2),                    # Draw position
                        (Room.CANVAS_WIDTH, Room.CANVAS_HEIGHT))                            # Scale to fit
        
        # Draw Play button
        canvas.draw_polygon([(self.PLAY_BUTTON_X, self.PLAY_BUTTON_Y),
                            (self.PLAY_BUTTON_X + self.PLAY_BUTTON_WIDTH, self.PLAY_BUTTON_Y),
                            (self.PLAY_BUTTON_X + self.PLAY_BUTTON_WIDTH, self.PLAY_BUTTON_Y + self.PLAY_BUTTON_HEIGHT),
                            (self.PLAY_BUTTON_X, self.PLAY_BUTTON_Y + self.PLAY_BUTTON_HEIGHT)],
                            2, "White", "Gray"
                            )

        canvas.draw_text("PLAY AGAIN", (self.PLAY_BUTTON_X + 8, self.PLAY_BUTTON_Y + 25), 15, "White")
        canvas.draw_text(f"Score : {self.score}", (self.PLAY_BUTTON_X + 10, self.PLAY_BUTTON_Y + 100), 20, "White")

    def mouse_click(self, pos):
        """Handles mouse clicks and starts the game if Play button is clicked."""
        if (self.PLAY_BUTTON_X <= pos[0] <= self.PLAY_BUTTON_X + self.PLAY_BUTTON_WIDTH and
                self.PLAY_BUTTON_Y <= pos[1] <= self.PLAY_BUTTON_Y + self.PLAY_BUTTON_HEIGHT):
            if self.state == "menu":
                self.state = "game"
            elif self.state == "finished":
                self.reset_game()