try:
    import simplegui
except ImportError:
    import SimpleGUICS2Pygame.simpleguics2pygame as simplegui

import random

from Entity import Obstacle, Enemy, BossEnemy, Player, Item
from Vector import Vector

# Room file is for generating rooms for the player to traverse
# Room Constants in Room
# Canvas Constants in Room
# Door Constants in Door

class Room:
    """A single room in the game world with walls, obstacles, enemies, and a door.
    
    Rooms contain walls, obstacles, enemies, and a door.
    The player must defeat all enemies to unlock the door and progress.
    
    Attributes:
        walls (list): List of wall obstacles
        obstacles (list): List of obstacle objects
        enemies (list): List of enemy entities
        door (list): Door object
            * Data type is a list because I tried making multiple doors but it became a lot more complex.
             I should change it to just take a single Door object, I'll get around to it later.
    """

    # Room constants
    WALL_THICKNESS = 20

    # Canvas size
    CANVAS_WIDTH = 800
    CANVAS_HEIGHT = 600


    def __init__(self):
        self.walls = []
        self.door = []
        self.initialize_room()
    
    def initialize_room(self):
        """Generates the room layout including walls, obstacles, and enemies."""

        # Make using constants easier using these instead
        wall_thickness = self.WALL_THICKNESS
        canvas_width = self.CANVAS_WIDTH
        canvas_height = self.CANVAS_HEIGHT
        door_width = Door.DOOR_WIDTH
        
        # Create east, west, south walls
        self.walls = [
            Obstacle(Vector(canvas_width / 2, canvas_height - wall_thickness / 2), canvas_width, wall_thickness),  # Bottom wall
            Obstacle(Vector(wall_thickness / 2, canvas_height / 2), wall_thickness, canvas_height),  # Left wall
            Obstacle(Vector(canvas_width - wall_thickness / 2, canvas_height / 2), wall_thickness, canvas_height)  # Right wall
        ]

        # Create top wall with gap for door
        top_door_left = (canvas_width / 2) - (door_width / 2)
        top_door_right = (canvas_width / 2) + (door_width / 2)

        if top_door_left > 0:
            self.walls.append(Obstacle(Vector(top_door_left / 2, wall_thickness / 2), top_door_left, wall_thickness))

        if top_door_right < canvas_width:
            self.walls.append(Obstacle(Vector(top_door_right + (canvas_width - top_door_right) / 2, wall_thickness / 2),
                                       canvas_width - top_door_right, wall_thickness))

        # Create door at the top of the room
        self.door = [Door(Vector(canvas_width / 2, 0))]
    
    def _generate_obstacles(self):

        # Make using constants easier using these instead
        wall_thickness = self.WALL_THICKNESS
        canvas_width = self.CANVAS_WIDTH
        canvas_height = self.CANVAS_HEIGHT

        num_obstacles = random.randint(*self.OBSTACLE_COUNT_RANGE)
        for _ in range(num_obstacles):
            valid_position = False
            while not valid_position:
                width = random.randint(30, 80)
                height = random.randint(30, 80)
                pos = Vector(
                    random.randint(wall_thickness + width//2, canvas_width - wall_thickness - width//2),
                    random.randint(wall_thickness + height//2, canvas_height - wall_thickness - height//2)
                )
                
                # Make sure obstacles don't block doors
                valid_position = True
                for door in self.door:
                    if (door.position - pos).length() < 100:
                        valid_position = False
                        break
                
                if valid_position:
                    self.obstacles.append(Obstacle(pos, width, height))
    
    def _generate_enemies(self):
        # Make using constants easier with these
        wall_thickness = self.WALL_THICKNESS
        canvas_width = self.CANVAS_WIDTH
        canvas_height = self.CANVAS_HEIGHT
        enemy_radius = Enemy.ENEMY_RADIUS

        num_enemies = random.randint(*self.ENEMY_COUNT_RANGE)
        for _ in range(num_enemies):
            valid_position = False
            while not valid_position:
                pos = Vector(
                    random.randint(wall_thickness + enemy_radius, canvas_width - wall_thickness - enemy_radius),
                    random.randint(wall_thickness + enemy_radius, canvas_height - wall_thickness - enemy_radius)
                )
                
                # Make sure enemies don't spawn in obstacles
                valid_position = True
                for obstacle in self.obstacles:
                    if obstacle.collides_with_circle(pos, enemy_radius * 2):
                        valid_position = False
                        break
                
                # Make sure enemies don't spawn near doors
                for door in self.door:
                    if (door.position - pos).length() < 100:
                        valid_position = False
                        break
                
                if valid_position:
                    self.enemies.append(Enemy(pos))
    
    def update(self, player, list_of_entities, interaction):
        """Update all entities in the room and handle interactions."""

        # Make using constants easier using these
        player_attack_radius = player.attack_radius
        player_attack_damage = player.attack_damage
        
        entities = list_of_entities

        for entity in entities[:]:
            if isinstance(entity, Enemy) or isinstance(entity, BossEnemy):
                entity.update(player.position)
            else:
                entity.update()
            
            self.handle_collisions(entity)
            
            if player.is_attacking:
                attack_pos = player.position + player.direction.get_normalized() * player_attack_radius
                if (attack_pos - entity.position).length() < player.attack_radius + entity.radius:
                    entity.take_damage(player_attack_damage)
                    
            if not entity.alive:
                entities.remove(entity)
                if isinstance(entity, Enemy):
                    interaction.score += 100
                if isinstance(entity, Item):
                    entity.add_powerUp(player)
        
        if len(entities) == 0:
            self.door[0].set_open(True)
    
    def handle_collisions(self, entity):
        """Handle collisions between an entity and room elements.
        
        Prevents entities from passing through walls, obstacles and closed doors.
        
        Args:
            entity (Entity): The entity to check collisions for
        """
        # Check collisions with walls
        for wall in self.walls:
            if wall.collides_with_circle(entity.position, entity.radius):
                # Find closest point on wall boundary
                closest_x = max(wall.bounds["left"], min(entity.position.x, wall.bounds["right"]))
                closest_y = max(wall.bounds["top"], min(entity.position.y, wall.bounds["bottom"]))
                closest_point = Vector(closest_x, closest_y)
                
                # Calculate direction from wall to entity
                direction = entity.position - closest_point
                
                # Hacky alternative way to handle sticky problem
                if direction.length() < 0.001:
                    direction = Vector(1, 1)
                
                # Normalize and scale by radius to move entity outside wall
                direction = direction.get_normalized().multiply(entity.radius + 1)
                entity.position = closest_point + direction
                
                # Stop movement in collision direction
                if abs(direction.x) > abs(direction.y):
                    entity.velocity.x = 0
                else:
                    entity.velocity.y = 0
        
        # Check collisions with doors
        for door in self.door:
            if not door.is_open and isinstance(entity, Player):
                if (abs(entity.position.x - door.position.x) < door.width/2 and 
                    entity.position.y < entity.radius * 2):
                    entity.position.y = entity.radius * 2
                    entity.velocity.y = 0
                
    def draw(self, canvas):
        for wall in self.walls:
            wall.draw(canvas)
        
        for door in self.door:
            door.draw(canvas)
    
    def check_door_entered(self, player):
        """Check if the player has entered the door.
            
        Returns:
            int: 1 if entered, -1 otherwise
        """
        if self.door[0].check_entry(player):
            return 1
        return -1
    
class Door:
    """Door that allows transition between rooms.
    
    Doors can be open or closed based on room state (enemies defeated).
    
    Attributes:
        position (Vector): The position of the door
        is_open (bool): Whether the door is open
        width (float): The width of the door
        height (float): The height of the door
        color (str): The current color based on open/closed state
    """

    # Door Constants
    DOOR_WIDTH = 60
    DOOR_HEIGHT = 20
    DOOR_COLOR_CLOSED = "Brown"
    DOOR_COLOR_OPEN = "Green"

    def __init__(self, position):
        self.position = position
        self.is_open = False
        self.width = self.DOOR_WIDTH
        self.height = self.DOOR_HEIGHT
        self.color = self.DOOR_COLOR_CLOSED
    
    def set_open(self, is_open):
        """Set the door's open state and update its color.
        
        Args:
            is_open (bool): Whether the door should be open
        """
        self.is_open = is_open
        self.color = self.DOOR_COLOR_OPEN if is_open else self.DOOR_COLOR_CLOSED
    
    def draw(self, canvas):
        x = self.position.x - self.width/2
        y = 0
        canvas.draw_polygon([
            (x, y), 
            (x + self.width, y), 
            (x + self.width, y + self.height), 
            (x, y + self.height)
        ], 1, "Black", self.color)
    
    def check_entry(self, player):
        """Check if the player is entering through this door.
        
        Returns:
            bool: True if player is entering, False otherwise
        """
        is_at_door = (abs(player.position.x - self.position.x) < self.width/2 and 
            player.position.y < player.radius)
        
        if self.is_open and is_at_door:
            return True
        
        return False

class EnemyRoom(Room):

    # Enemy Room Constants
    OBSTACLE_COUNT_RANGE = (3, 7)  # min and max number of obstacles per room
    ENEMY_COUNT_RANGE = (3, 6)  # min and max number of enemies per room

    def __init__(self, room_count):
        self.obstacles = []
        self.enemies = []
        # Use room count to increase the amount of possible enemies
        self.room_number = room_count
        self.ENEMY_COUNT_RANGE = (3 + (self.room_number // 2), 6 + (self.room_number // 2))
        super().__init__()

    def initialize_room(self):
        super().initialize_room()
        self._generate_obstacles()
        self._generate_enemies()

    def update(self, player, interaction):
        super().update(player, self.enemies, interaction)

        for enemy in self.enemies:
            enemy.attack(player)

    def handle_collisions(self, entity):
        super().handle_collisions(entity)

        # Check collisions with obstacles
        for obstacle in self.obstacles:
            if obstacle.collides_with_circle(entity.position, entity.radius):
                closest_x = max(obstacle.bounds["left"], min(entity.position.x, obstacle.bounds["right"]))
                closest_y = max(obstacle.bounds["top"], min(entity.position.y, obstacle.bounds["bottom"]))
                closest_point = Vector(closest_x, closest_y)
                
                direction = entity.position - closest_point
                
                if direction.length() < 0.001:
                    direction = Vector(1, 1)
                
                direction = direction.get_normalized().multiply(entity.radius + 1)
                entity.position = closest_point + direction
                
                if abs(direction.x) > abs(direction.y):
                    entity.velocity.x = 0
                else:
                    entity.velocity.y = 0

    def draw(self, canvas):
        super().draw(canvas)
        for obstacle in self.obstacles:
            obstacle.draw(canvas)
        
        for enemy in self.enemies:
            enemy.draw(canvas)

class ItemRoom(Room):

    # Item Room Constants
    NUM_OF_ITEMS = 3

    def __init__(self):
        # Create 3 items
        self.items = [Item(1), Item(2), Item(3)]
        super().__init__()

    def update(self, player, interaction):
        """Update all entities in the room and handle interactions."""

        super().update(player, self.items, interaction)  

        # If one item taken, get rid of rest
        if len(self.items) < 3:
            for item in self.items:
                self.items.remove(item)
        
        # Door always open, can leave without taking item
        self.door[0].set_open(True)

    def draw(self, canvas):
        super().draw(canvas)
        for item in self.items:
            item.draw(canvas)

class BossRoom(Room):
    def __init__(self, room_count):
        self.room_number = room_count
        self.bosses = [BossEnemy(Vector(self.CANVAS_WIDTH / 2, self.CANVAS_HEIGHT / 2))]
        super().__init__()

    def update(self, player, interaction):
        super().update(player, self.bosses, interaction)

        for boss in self.bosses:
            boss.attack(player)

    def draw(self, canvas):
        super().draw(canvas)
        
        for boss in self.bosses:
            boss.draw(canvas)